
#include "interfaces/irdb/STARSOp.h"
#include "interfaces/STARSTypes.h"
#include "interfaces/SMPDBInterface.h"
#include "interfaces/irdb/STARSProgram.h"
#include "interfaces/irdb/STARSInstruction.h"

#include <libIRDB-core.hpp>
#include <iostream>


using namespace std;

static int log2int(unsigned int index)
{
	assert(index>=0);
	int targetlevel = 0;
	while (index >>= 1) ++targetlevel;
	return  targetlevel;
}


static int log2int_or_err(unsigned int index)
{
	if(index==0)
		return -1;
	return log2int(index);
}


STARS_IRDB_op_t::STARS_IRDB_op_t(const DISASM &d, int indx, const ARGTYPE &the_arg, const int length)
{
	Init(); // make void
	int bea_regno=bea_regno=log2int_or_err(the_arg.ArgType&0xFFFF);
	assert(length>=0 && length<=16); // x86-insns are 1-16 bytes, but we allow 0 in case you're lazy.

	visible=1;
	switch((the_arg.ArgType&0xFFFF0000))
	{
  		case NO_ARGUMENT:
			/* leave as void */
			break;
  		case REGISTER_TYPE + GENERAL_REG :
		{
			OpType=op_Reg;

			     if(string(the_arg.ArgMnemonic)=="al")  { operand.reg.RegNum=STARS_x86_R_al; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="cl")  { operand.reg.RegNum=STARS_x86_R_cl; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="dl")  { operand.reg.RegNum=STARS_x86_R_dl; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="bl")  { operand.reg.RegNum=STARS_x86_R_bl; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="ah")  { operand.reg.RegNum=STARS_x86_R_ah; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="ch")  { operand.reg.RegNum=STARS_x86_R_ch; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="dh")  { operand.reg.RegNum=STARS_x86_R_dh; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="bh")  { operand.reg.RegNum=STARS_x86_R_bh; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="spl")  { operand.reg.RegNum=STARS_x86_R_spl; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="spl")  { operand.reg.RegNum=STARS_x86_R_spl; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="bpl")  { operand.reg.RegNum=STARS_x86_R_bpl; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="sil")  { operand.reg.RegNum=STARS_x86_R_sil; byteWidth=1; }
			else if(string(the_arg.ArgMnemonic)=="dil")  { operand.reg.RegNum=STARS_x86_R_dil; byteWidth=1; }
			else 
			{
				// set the operand.reg.RegNum appropriately.
				operand.reg.RegNum=(STARS_RegNo)(STARS_x86_R_ax+bea_regno);		// regno may be -1, but ax-1==none, which is right.

				// if the first character is an r, it's a 64-bit reg.
				// e.g. rax, rsp, rdi; UNLESS there is a suffix
				if (the_arg.ArgMnemonic[0] == 'r') {
					byteWidth = 8;
					size_t ArgNameLen = strlen(the_arg.ArgMnemonic);
					char Suffix = (char) tolower(the_arg.ArgMnemonic[ArgNameLen - 1]);
					if ('b' == Suffix)
						byteWidth = 1;
					else if ('w' == Suffix)
						byteWidth = 2;
					else if ('d' == Suffix)
						byteWidth = 4;
				}

				// otherwise, if the first character is an e, and the 3rd char is an 'x', 'p', or 'i'  it's a 32-bit reg. 
				// e.g., eax, esp, edi
				// caution: es
				else if(the_arg.ArgMnemonic[0]=='e' && (the_arg.ArgMnemonic[2]=='x' || the_arg.ArgMnemonic[2]=='i' || the_arg.ArgMnemonic[2]=='p'))
					byteWidth=4;
				// e.g	ax, sp, bp, di
				else if(the_arg.ArgMnemonic[1]=='x' || the_arg.ArgMnemonic[1]=='i' || the_arg.ArgMnemonic[1]=='p')
					byteWidth=2;
				else 
					byteWidth=1;
			}
			break;
		}
  		case REGISTER_TYPE + MMX_REG :
			byteWidth=8;
			OpType=op_MMXReg;
			assert(bea_regno!=-1);
			operand.reg.RegNum=(STARS_RegNo)(STARS_x86_R_mm0+bea_regno);
			break;
  		case REGISTER_TYPE + FPU_REG :
			byteWidth=8;
			OpType=op_MMXReg;
			assert(bea_regno!=-1);
			operand.reg.RegNum=(STARS_RegNo)(STARS_x86_R_st0+bea_regno);
			break;
  		case REGISTER_TYPE + SSE_REG :
			if(the_arg.ArgMnemonic[0]=='x')
			{
				byteWidth=8;
				OpType=op_XMMReg;
				assert(bea_regno!=-1);
				operand.reg.RegNum=(STARS_RegNo)(STARS_x86_R_xmm0+bea_regno);
				break;
			}
			else if(the_arg.ArgMnemonic[0]=='y')
			{
				byteWidth=16;
				OpType=op_YMMReg;
				assert(bea_regno!=-1);
				operand.reg.RegNum=(STARS_RegNo)(STARS_x86_R_ymm0+bea_regno);
				break;
			}
			else
				assert(0);
  		case REGISTER_TYPE + CR_REG :
			assert(0);
/*
no stars type for control regs?
			OpType=CrReg;
			operand.reg.RegNum=STARS_x86_R_cr0+bea_regno;
*/
			break;
  		case REGISTER_TYPE + DR_REG :
			assert(0);
/*
no stars type for debug regs?
			OpType=op_CrReg;
			operand.reg.RegNum=STARS_x86_R_cr0+bea_regno;
*/
			break;
  		case REGISTER_TYPE + SPECIAL_REG :
		{
			assert(0);
/*

this isn't even close to right.  should be gdtr, ldtr, idtr, tr, etc.
no operands for eflags or mxcsr?
			switch(bea_regno)
			{
				case EFLAGS:
					OpType=op_Eflags;
					operand.reg.RegNum=0;
					break;
				case MXCSR:
					OpType=op_Mxcsr;
					operand.reg.RegNum=0;
					break;
					break;
				assert(0);
			}
			break;
*/
		}
  		case REGISTER_TYPE + MEMORY_MANAGEMENT_REG :
			assert(0);
			break;
  		case REGISTER_TYPE + SEGMENT_REG :
			OpType=op_SegReg;
			assert(bea_regno!=-1);
			operand.reg.RegNum=(STARS_RegNo)(STARS_x86_R_es+bea_regno);
			break;
  		case MEMORY_TYPE:
  		case MEMORY_TYPE + RELATIVE_:
  		case MEMORY_TYPE + ABSOLUTE_:
		{
			// log2int_or_err returns -1 if the value is 0 (since you can't log2(0)
			// which maps nicely to STARS_NN_None for no-register.
			operand.mem.base=(STARS_RegNo)log2int_or_err(the_arg.Memory.BaseRegister);
			operand.mem.index=(STARS_RegNo)log2int_or_err(the_arg.Memory.IndexRegister);
			operand.mem.scale=the_arg.Memory.Scale;
			if( (the_arg.ArgType&0xFFFF0000) == (MEMORY_TYPE + RELATIVE_))
			{
				assert(length!=0);// we need the length in this case.
				operand.mem.disp=the_arg.Memory.Displacement+length;
				operand.mem.base=STARS_x86_R_ip;
			}
			else
			{
				operand.mem.disp=the_arg.Memory.Displacement;
			}


			// if there's no base reg, and an index reg with scale==1,
			// ida represents this oddly as a base with no sib.
			if(operand.mem.base ==  STARS_x86_R_none && operand.mem.scale==1 && operand.mem.index!=STARS_x86_R_none)
			{
				// change representation.
				operand.mem.base=operand.mem.index;
				operand.mem.index=STARS_x86_R_none;
				operand.mem.scale=0;
			}

			// if the scale >0, we have a sib byte.
			if(operand.mem.scale > 0)
			{
				operand.mem.hasSIB=1;
				operand.mem.scale=log2int(operand.mem.scale);	
			}
			// if there's both a base and an index, then there's a sib byte.
			if(operand.mem.base != STARS_x86_R_none && operand.mem.index != STARS_x86_R_none)
				operand.mem.hasSIB=1;


			byteWidth=the_arg.ArgSize/8;
			OpType=op_Mem;
			break;
		}
  		case CONSTANT_TYPE + RELATIVE_ :
			OpType=op_Imm;
			operand.imm.imm=d.Instruction.Immediat;
			operand.imm.pc_rel=true;
			byteWidth=the_arg.ArgSize/8;
			break;
  		case CONSTANT_TYPE + ABSOLUTE_ :
			OpType=op_Imm;
			switch(the_arg.ArgSize)
			{
				case 8:
					operand.imm.imm=(signed int)(signed char)d.Instruction.Immediat;
					break;
				case 16:
					operand.imm.imm=(signed int)(signed short)d.Instruction.Immediat;
					break;
				case 32:
				default:
					operand.imm.imm=d.Instruction.Immediat;
					break;
		
			}
			operand.imm.pc_rel=false;
			byteWidth=the_arg.ArgSize/8;
			break;
		default:
			assert(0);
	}


	// deal with segment register
	switch(the_arg.SegmentReg)
	{
		case 0:		// bea reporting 0 means no register.
			SegReg=STARS_x86_R_none;	// map to no reg.
			break;
		case ESReg:	// bea seg reg ESReg maps to 
        		SegReg=STARS_x86_R_es;    // STARS_x86_R_es reg.
			break;
		case CSReg:	// etc.
        		SegReg=STARS_x86_R_cs;
			break;
		case SSReg:
        		SegReg=STARS_x86_R_ss;
			break;
		case DSReg:
        		SegReg=STARS_x86_R_ds;
			break;
		case FSReg:
        		SegReg=STARS_x86_R_fs;
			break;
		case GSReg:
        		SegReg=STARS_x86_R_gs;
			break;
		default:
			assert(0);

	}






}

bool STARS_IRDB_op_t::operator<(const STARS_op_t &rOp_param) const
{
	const STARS_IRDB_op_t* p=dynamic_cast<const STARS_IRDB_op_t*>(&rOp_param);
	assert(p);
	const STARS_IRDB_op_t& rOp=*p;

	if(GetOpType() != rOp.GetOpType())
		return GetOpType() < rOp.GetOpType();

	switch(GetOpType())
	{
		case op_Mem: 
			if ( operand.mem.base != rOp.operand.mem.base)
				return ( operand.mem.base < rOp.operand.mem.base);
			if ( operand.mem.index != rOp.operand.mem.index)
				return ( operand.mem.index < rOp.operand.mem.index);
			if ( operand.mem.scale != rOp.operand.mem.scale)
				return ( operand.mem.scale < rOp.operand.mem.scale);
			if ( operand.mem.disp != rOp.operand.mem.disp)
				return ( operand.mem.disp < rOp.operand.mem.disp);
			return false; // equal
		case op_Imm: 
			return operand.imm.imm < rOp.operand.imm.imm;
		case op_CrReg: 
		case op_DrReg: 
		case op_Eflags: 
		case op_MMXReg: 
		case op_XMMReg: 
		case op_YMMReg: 
		case op_FPReg: 
		case op_Reg: 
			return operand.reg.RegNum < rOp.operand.reg.RegNum;

		case op_Void: 
		case op_Mxcsr: 
		case op_SegReg:
			return true;

		case op_Addr:
			return operand.addr.addr < rOp.operand.addr.addr;

		default:
			assert(0);
			
	}
}

char STARS_IRDB_op_t::GetSIB(void) const
{
	char ret=0;
	ret|=((int)operand.mem.base&0x7)<<0;
	ret|=((int)operand.mem.index&0x7)<<3;
	ret|=(((int)operand.mem.scale)&0x3)<<6;
	return ret;
} // Get x86 SIB byte (dense encoding of base reg, index reg, and scale factor)


void STARS_IRDB_op_t::SetSIB(char value)
{
	assert(OpType==op_Mem);
	operand.mem.hasSIB=true;

	// update the base register.  Odd casting to avoid compiler error bcause base is a STARS_RegNo
	(*(int*)&operand.mem.base)&=~0x7;
	(*(int*)&operand.mem.base)|=(value&0x7);

	// Set scale.  Odd casting to avoid compiler error bcause base is a STARS_RegNo
	(*(int*)&operand.mem.index)&=~(0x7<<3);
	(*(int*)&operand.mem.index)|=(value>>3)&0x7;

	// set scale
	operand.mem.scale= (STARS_RegNo)((value>>6)&0x3);
	if(operand.mem.index==STARS_x86_R_sp)
		operand.mem.scale=(STARS_RegNo)0;

	assert(value==GetSIB());
}

void STARS_IRDB_op_t::Dump()
{

	cout<<"Dumping operand at "<<hex<<this<<endl;
	cout<<"Byte width:"<<dec<<byteWidth<<endl;
	if(visible)
		cout<<"Operand is visible"<<endl;
	else 
		cout<<"Operand is NOT visible"<<endl;

	switch(OpType)
	{

                case op_Reg:
			cout<<"Type: Reg, regno="<<dec<<GetReg()<<endl;
			break;
		case op_Mem:
			cout<<"Type: Mem["<<MDGetSIBBaseReg()<<"+"<<MDGetSIBIndexReg()<<"*"
			    <<GetSIBScaleFactor()<<"+"<<hex<<GetAddr()<<"]"<<endl;
			break;
		case op_Imm:
			cout<<"Type: Imm:"<<hex<<GetAddr()<<endl;
			break;
		case op_Void:
			cout<<"Type: Void"<<endl;
			break;
		case op_CrReg:
			cout<<"Type: CrReg, regno="<<dec<<GetReg()<<endl;
			break;
		case op_DrReg:
			cout<<"Type: DrReg, regno="<<dec<<GetReg()<<endl;
			break;
		case op_Eflags:
			cout<<"Type: Eflags"<<endl;
			break;
		case op_MMXReg:
			cout<<"Type: MmxReg, regno="<<dec<<GetReg()<<endl;
			break;
		case op_XMMReg:
			cout<<"Type: XmmReg, regno="<<dec<<GetReg()<<endl;
			break;
		case op_YMMReg:
			cout<<"Type: YmmReg, regno="<<dec<<GetReg()<<endl;
			break;
		case op_FPReg:
			cout<<"Type: FpReg, regno="<<dec<<GetReg()<<endl;
			break;
		case op_Mxcsr:
			cout<<"Type: Mxcsr"<<endl;
			break;
		case op_SegReg:
			cout<<"Type: SegReg, regno="<<dec<<GetReg()<<endl;
			break;
		case op_Addr:
			cout<<"Type: Addr:"<<hex<<GetAddr()<<endl;
			break;

		
	}
	
}


uint16_t STARS_IRDB_op_t::GetByteWidth(void) const
{
	if(OpType==op_Void)
		return (uint16_t) global_STARS_program->GetSTARS_ISA_Bytewidth();
	else
		return byteWidth;
}

